import numpy as np
import pytest

from pytest_pl_grader.fixture import FeedbackFixture
from pytest_pl_grader.fixture import StudentFixture


@pytest.mark.grading_data(name="test_query_func", points=2)
def test_query_func(sandbox: StudentFixture) -> None:
    assert sandbox.query_function("fib", 3) == 5


def test_numpy(sandbox: StudentFixture) -> None:
    expected_res = np.array([1.0, 2.0, 3.0, 4.0, 5.0])
    np.testing.assert_array_equal(sandbox.query("my_array"), expected_res)


def test_temp(sandbox: StudentFixture, feedback: FeedbackFixture) -> None:
    # This is a test to check if the sandbox fixture is working
    assert sandbox is not None

    assert sandbox.query("x") == 5
    feedback.add_message("This is a test message")


def test_temp_2(sandbox: StudentFixture, feedback: FeedbackFixture) -> None:
    # This is a test to check if the sandbox fixture is working
    assert sandbox is not None

    if sandbox.query("x") == 5:
        feedback.set_score(0.5)


"""
def test_help(pytester):
    result = pytester.runpytest_subprocess("--help")
    result.stdout.fnmatch_lines(
        [
            "*",
            "*",
            "benchmark:",
            "  --benchmark-min-time=SECONDS",
            "                        *Default: '0.000005'",
            "  --benchmark-max-time=SECONDS",
            "                        *Default: '1.0'",
            "  --benchmark-min-rounds=NUM",
            "                        *Default: 5",
            "  --benchmark-timer=FUNC",
            "  --benchmark-calibration-precision=NUM",
            "                        *Default: 10",
            "  --benchmark-warmup=[KIND]",
            "  --benchmark-warmup-iterations=NUM",
            "                        *Default: 100000",
            "  --benchmark-disable-gc",
            "  --benchmark-skip      *",
            "  --benchmark-only      *",
            "  --benchmark-save=NAME",
            "  --benchmark-autosave  *",
            "  --benchmark-save-data",
            "  --benchmark-json=PATH",
            "  --benchmark-compare=[NUM|_ID]",
            "  --benchmark-compare-fail=EXPR?[[]EXPR?...[]]",
            "  --benchmark-cprofile=COLUMN",
            "  --benchmark-storage=URI",
            "                        *Default: 'file://./.benchmarks'.",
            "  --benchmark-verbose   *",
            "  --benchmark-sort=COL  *",
            "  --benchmark-group-by=LABEL",
            "                        *Default: 'group'",
            "  --benchmark-columns=LABELS",
            "  --benchmark-histogram=[FILENAME-PREFIX]",
            "*",
        ]
    )


def test_groups(testdir):
    test = testdir.makepyfile(

import time
import pytest

def test_fast(benchmark):
    benchmark(lambda: time.sleep(0.000001))
    assert 1 == 1

def test_slow(benchmark):
    benchmark(lambda: time.sleep(0.001))
    assert 1 == 1

@pytest.mark.benchmark(group="A")
def test_slower(benchmark):
    benchmark(lambda: time.sleep(0.01))
    assert 1 == 1

@pytest.mark.benchmark(group="A", warmup=True)
def test_xfast(benchmark):
    benchmark(lambda: None)
    assert 1 == 1

    )
    result = testdir.runpytest_subprocess("-vv", "--doctest-modules", test)
    result.stdout.fnmatch_lines(
        [
            "*collected 5 items",
            "*",
            "test_groups.py::*test_groups PASSED*",
            "test_groups.py::test_fast PASSED*",
            "test_groups.py::test_slow PASSED*",
            "test_groups.py::test_slower PASSED*",
            "test_groups.py::test_xfast PASSED*",
            "*",
            "* benchmark: 2 tests *",
            "*",
            "* benchmark 'A': 2 tests *",
            "*",
            "*====== 5 passed * ======*",
        ]
    )
    print(result.stdout)
    assert False
"""



import numpy.linalg as la


def most_consistent_stock(stock_returns, market_trend, p):
    index = p_norm = -1
    min_val = float("inf")

    for i, r in enumerate(stock_returns):
        pp = la.norm(r - market_trend, ord=p) / la.norm(r, ord=p)
        if pp < min_val:
            index = i
            min_val = pp
            p_norm = la.norm(r - market_trend, ord=p)

    return index, p_norm



import numpy as np
import pytest

from pytest_pl_grader.fixture import StudentFixture

### Start of simulation code ###


def simulate_stock_price(
    S_0,
    mu_0,
    sigma_0,
    N,
    T=1,
    kappa=1.5,
    theta=0.2,
    eta=0.3,
    seed=42,
) -> np.ndarray:
    """
    Simulates a realistic stock price time series with stochastic drift and
    volatility.

    Parameters:
        S_0: Initial stock price
        mu_0: Initial drift (expected return)
        sigma_0: Initial volatility
        N: Number of time steps
        T: Total time period (in years)
        kappa: Mean reversion speed of volatility
        theta: Long-term average volatility
        eta: Volatility of volatility (vol of vol)

    Returns:
        t: Time points
        S: Simulated stock prices
    """
    dt = T / N  # Time step
    # t = np.linspace(start=0, stop=T, num=N)  # Time grid

    # Initialize stock price, drift, and volatility arrays
    S = np.zeros(shape=N, dtype=np.float64)
    mu = np.zeros(shape=N, dtype=np.float64)
    sigma = np.zeros(shape=N, dtype=np.float64)
    np.random.seed(seed)  # For reproducibility
    W_stock = np.random.standard_normal(size=N)  # Random component for stock price
    W_vol = np.random.standard_normal(size=N)  # Random component for volatility

    # Set initial values
    S[0] = S_0
    mu[0] = mu_0
    sigma[0] = sigma_0

    for i in range(1, N):
        # Stochastic drift: Random walk with slight mean-reversion
        mu[i] = mu[i - 1] + np.random.normal(loc=0, scale=0.005)  # Small changes in drift

        # Stochastic volatility: Heston model (mean-reverting)
        sigma[i] = sigma[i - 1] + kappa * (theta - sigma[i - 1]) * dt + eta * np.sqrt(dt) * W_vol[i]
        sigma[i] = np.maximum(sigma[i], 0.01)  # Ensure positive volatility

        # Stock price follows modified GBM with changing mu and sigma
        S[i] = S[i - 1] * np.exp((mu[i] - 0.5 * sigma[i] ** 2) * dt + sigma[i] * np.sqrt(dt) * W_stock[i])

    return S


def add_noise(
    stock_prices,
    base_vol=0.01,
    vol_spike_prob=0.1,
    spike_factor=3,
    jump_prob=0.02,
    jump_magnitude=0.05,
    seed=42,
) -> np.ndarray:
    N = len(stock_prices)

    np.random.seed(seed)  # For reproducibility
    # Step 1: Gaussian Noise (Micro fluctuations)
    noise = np.random.normal(loc=0, scale=base_vol, size=N) * stock_prices

    # Step 2: GARCH-Like Volatility Clustering (Random High/Low Volatility Phases)
    spikes = np.random.rand(N) < vol_spike_prob  # Boolean array for volatility spikes
    noise[spikes] *= spike_factor  # Increase volatility in some places

    # Step 3: Jump Diffusion (Sudden Large Moves)
    jumps = np.random.rand(N) < jump_prob  # Boolean array for jump occurrences
    jump_size = np.random.uniform(low=-jump_magnitude, high=jump_magnitude, size=N) * stock_prices
    noise += jumps * jump_size  # Apply jumps where they occur

    # Final Noisy Prices
    stock_prices += noise

    return stock_prices


def simulate_stock_returns(S_0, mu_0, sigma_0, D, N):
    stock_returns = np.array(
        object=[simulate_stock_price(S_0=S_0, mu_0=mu_0, sigma_0=sigma_0, N=D) for _ in range(N)],
        dtype=np.float64,
    )

    market_trend = np.mean(a=stock_returns, axis=0)  # Average across stocks
    market_trend = add_noise(stock_prices=market_trend)

    for i in range(stock_returns.shape[0]):
        stock_returns[i] = add_noise(stock_prices=stock_returns[i])

    return stock_returns, market_trend


### End of simulation code ###


def get_starting_parameters():
    # Parameters
    S_0 = 100  # Initial stock price
    mu_0 = 0.05  # Initial expected return (5%)
    sigma_0 = 0.2  # Initial volatility (20%)
    D = 100  # Number of time steps (daily for 1 year)

    # TODO increase N to 1000 for more realistic data
    N = 10

    # Simulate stock market
    stock_returns, market_trend = simulate_stock_returns(S_0=S_0, mu_0=mu_0, sigma_0=sigma_0, D=D, N=N)
    np.random.seed(42)  # For reproducibility
    p = np.random.choice(a=[1, 3, 4, 5, 6, 7, 8, 9, 10])

    return stock_returns, market_trend, p


def most_consistent_stock(stock_returns, market_trend, p):
    dists = np.array([np.linalg.norm(stock - market_trend, ord=p) for stock in stock_returns])
    index = np.argmin(dists)
    p_norm = dists[index]

    return index, p_norm


@pytest.mark.grading_data(name="index", points=1)
def test_0(sandbox: StudentFixture) -> None:
    stock_returns, market_trend, p = get_starting_parameters()

    correct_index, _ = most_consistent_stock(stock_returns, market_trend, p)
    student_index, _ = sandbox.query_function("most_consistent_stock", stock_returns, market_trend, p)

    assert correct_index == student_index, "The index of the most consistent stock is incorrect."


@pytest.mark.grading_data(name="p_norm", points=1)
def test_1(sandbox: StudentFixture) -> None:
    stock_returns, market_trend, p = get_starting_parameters()

    _, correct_p_norm = most_consistent_stock(stock_returns, market_trend, p)
    _, student_p_norm = sandbox.query_function("most_consistent_stock", stock_returns, market_trend, p)

    assert correct_p_norm == student_p_norm, "The p_norm of the most consistent stock is incorrect."



from typing import Any
from typing import cast

import numpy as np
import pandas as pd
import pytest

from pytest_pl_grader.json_utils import from_json
from pytest_pl_grader.json_utils import from_server_json
from pytest_pl_grader.json_utils import to_json
from pytest_pl_grader.utils import deserialize_object_unsafe
from pytest_pl_grader.utils import serialize_object_unsafe


def test_serialize_numpy_array() -> None:
    # Create a numpy array
    arr = np.array([1, 2, 3, 4, 5])

    # Serialize the numpy array
    serialized = serialize_object_unsafe(arr)

    # Deserialize the numpy array
    deserialized = cast(np.typing.ArrayLike, deserialize_object_unsafe(serialized))

    # Check if the original and deserialized arrays are equal
    assert np.array_equal(arr, deserialized)


@pytest.mark.parametrize("obj", [np.bool(True), np.int32(42), np.float64(3.14), complex(1, 2), True])
def test_serialize_json(obj: Any) -> None:
    # Serialize the object to JSON-compatible format
    json_compatible = to_json(obj)

    # Deserialize back to original object
    deserialized = from_json(json_compatible)

    np.testing.assert_equal(obj, deserialized)


@pytest.mark.parametrize(
    ("obj", "expected_result"),
    [
        (
            {
                "_type": "ndarray",
                "_value": [
                    [0.6838187162381367, 0.9687669728013233, 0.8739536562117622],
                    [0.6863101440077359, 0.18857604228294678, 0.036780253047014266],
                    [0.3518597494339264, 0.5156619713512136, 0.1474092805728887],
                ],
                "_dtype": "float64",
            },
            np.array([[0.68381872, 0.96876697, 0.87395366], [0.68631014, 0.18857604, 0.03678025], [0.35185975, 0.51566197, 0.14740928]]),
        ),
        (
            {"_type": "ndarray", "_value": [[64, 41, 40], [93, 75, 87], [27, 91, 27]], "_dtype": "int64"},
            np.array([[64, 41, 40], [93, 75, 87], [27, 91, 27]]),
        ),
    ],
)
def test_from_server_json_numpy(obj: dict, expected_result: np.ndarray) -> None:
    # Deserialize back to original object
    deserialized = from_server_json(obj)

    # Check if the original and deserialized arrays are equal
    np.testing.assert_allclose(deserialized, expected_result)


@pytest.mark.parametrize(
    ("obj", "expected_result"),
    [
        (
            {
                "_type": "dataframe",
                "_value": {
                    "index": [0, 1, 2, 3, 4],
                    "columns": ["ColA", "ColB", "ColC"],
                    "data": [[6, 98, 76], [91, 43, 6], [89, 46, 81], [48, 3, 19], [43, 31, 73]],
                },
            },
            pd.DataFrame(np.array([[6, 98, 76], [91, 43, 6], [89, 46, 81], [48, 3, 19], [43, 31, 73]]), columns=["ColA", "ColB", "ColC"]),
        )
    ],
)
def test_from_server_json(obj: dict, expected_result: pd.DataFrame) -> None:
    # Deserialize back to original object
    deserialized = from_server_json(obj)

    # Check if the original and deserialized arrays are equal
    pd.testing.assert_frame_equal(deserialized, expected_result)



import time

import numpy as np

x = 5


def fib(n):
    if n <= 1:
        return n + 1
    if n == 8:
        time.sleep(20)
    if n == 7:
        1 / 0

    return fib(n - 1) + fib(n - 2)


my_array = np.array([1.0, 2.0, 3.0, 4.0, 5.0])



import math
import platform
from collections import defaultdict
from pathlib import Path

import pytest

pytest_plugins = ("pytester",)
platform  # noqa: B018
import json

# --- Configuration ---
# The root directory where your scenario folders are located
SCENARIO_ROOT = Path(__file__).parent / "scenario_root"

# The file you want to copy into each scenario directory.
# This could be your autograder's conftest.py, or a specific test file.
# FILE_TO_COPY = Path(__file__).parent / "scenario_root" / "scenario.py"

CONVERSION_DICT = {
    "error": "errors",
    "passed": "passed",
    "failed": "failed",
    "skipped": "skipped",
}


# --- Helper Functions ---
def print_file_structure(path: Path, indent: str = "") -> None:
    """Recursively prints the file structure of a given directory."""
    print("-" * 100)

    for item in path.iterdir():
        print(indent + item.name)
        if item.is_dir():
            print_file_structure(item, indent + "  ")


def get_scenario_dirs(root_dir: Path):
    """
    Discovers all scenario directories within the given root_dir
    that start with 'test_'.
    """
    scenario_dirs = []
    for item in root_dir.iterdir():
        if item.is_dir() and item.name.startswith("test_"):
            scenario_dirs.append(item)

    return scenario_dirs


# --- Test Function with pytester ---
@pytest.mark.parametrize("scenario_dir", get_scenario_dirs(SCENARIO_ROOT), ids=lambda d: d.name)
def test_autograder_scenario_with_pytester(pytester: pytest.Pytester, scenario_dir: Path) -> None:
    """
    Tests each autograder scenario using the pytester fixture:
    1. Copies all files from the scenario directory (and the common testing file)
       into pytester's isolated test directory.
    2. Runs pytest within that isolated environment.
    3. Asserts on the pytest outcome (passed, failed, etc.) and captured stdout/stderr.
    """
    print(f"\n--- Running scenario: {scenario_dir.name} with pytester ---")

    # Ensure the common file to copy exists
    # if not FILE_TO_COPY.exists():
    #     pytest.fail(f"Error: FILE_TO_COPY '{FILE_TO_COPY}' does not exist.")

    # 1. Copy all files from the scenario directory into pytester's test directory
    # pytester.makepyfile() is for creating files with content,
    # pytester.copy_example() is good for copying directories or specific files.
    # We will iterate and copy all files from the scenario_dir and the common file.
    pytester_scenario_dir = pytester.mkdir(scenario_dir.name)

    new_file_name = scenario_dir.name + ".py"
    expected_outcome_dict = None

    # Copy files from the specific scenario directory
    for item in scenario_dir.iterdir():
        if item.is_file():
            if item.name == "scenario.py":
                # print("in the scenario.py branch", scenario_dir.name)
                pytester.path.joinpath(new_file_name).write_text(item.read_text())  # Copy the file content
            elif item.name == "expected_outcome.json":
                expected_outcome_dict = json.loads(item.read_text())
            elif item.name != "__init__.py":  # Ignore __init__.py files
                pytester_scenario_dir.joinpath(item.name).write_text(item.read_text())  # Copy the file content
                # print(f"Copied '{item.name}' to pytester's testdir.")
        elif item.is_dir():
            raise ValueError("Scenario directories should not contain subdirectories. Please flatten the structure.")

    if expected_outcome_dict is None:
        pytest.fail(f"Error: expected_outcome.json not found in scenario directory '{scenario_dir.name}'.")

    # You can pass additional arguments to pytest here, e.g., '-s' for print statements
    result = pytester.runpytest("-v", new_file_name)

    results_obj = json.loads((pytester.path / "autograder_results.json").read_text())

    """
    #NOTE uncomment this to save the results object to a file
    output_path = scenario_dir / "autograder_output.json"
    print(output_path)
    with open(output_path, "w") as f:
        json.dump(results_obj, f, indent=4, sort_keys=True)
    """

    expected_data_obj = expected_outcome_dict["expected_data_object"]

    assert math.isclose(expected_data_obj["score"], results_obj["score"])

    outcome_dict: defaultdict[str, int] = defaultdict(int)

    # TODO add tests for the tests object
    test_results_obj = {test_result["test_id"]: test_result for test_result in results_obj["tests"]}
    for expected_test in expected_data_obj["tests"]:
        test_id = expected_test["test_id"]

        # TODO make this not depend on the file name
        assert test_id in test_results_obj, f"Test '{test_id}' not found in results."
        actual_test = test_results_obj[test_id]

        # NOTE the message here is just a pattern that has to be contained in the actual message
        # this is to avoid issues with longer messages
        expected_message = expected_test.get("message")

        if expected_message is not None:
            assert expected_message in actual_test["message"], f"Message mismatch for test '{test_id}'."

        assert actual_test["max_points"] == expected_test["max_points"], f"Max points mismatch for test '{test_id}'."
        assert math.isclose(expected_test["points_frac"], actual_test["points_frac"]), f"Points fraction mismatch for test '{test_id}'."
        assert math.isclose(expected_test["points"], actual_test["points"]), f"Points mismatch for test '{test_id}'."
        assert actual_test["outcome"] == expected_test["outcome"], f"Outcome mismatch for test '{test_id}'."
        outcome_dict[CONVERSION_DICT[actual_test["outcome"]]] += 1

    result.assert_outcomes(**outcome_dict)

    # For a scenario where student code is expected to fail:
    # result.assert_outcomes(failed=1, passed=1) # If one test fails and one passes

    # You can also inspect stdout and stderr if your plugin prints output
    # result.stdout.fnmatch_lines(



import numpy as np

from pytest_pl_grader.fixture import StudentFixture


def test_random(sandbox: StudentFixture) -> None:
    # Check that variables we set are unaltered
    assert sandbox.query_setup("a") == 42
    assert sandbox.query_setup("b")["c"] == 1
    assert sandbox.query_setup("weird_func")(10) == 30

    assert sandbox.query("c") == 42
    assert sandbox.query("a") == 10
    assert sandbox.query("b")["c"] == 50
    assert sandbox.query("d") == 11

    sandbox.query("numpy_array")

    np.testing.assert_allclose(
        sandbox.query("numpy_array"),
        np.array([[0.68381872, 0.96876697, 0.87395366], [0.68631014, 0.18857604, 0.03678025], [0.35185975, 0.51566197, 0.14740928]]),
    )



def not_allowed(*args, **kwargs):
    raise RuntimeError("Calling sympy.series is not allowed in this question. You can still use sympy to compute derivatives")


import sympy as sp

sp.series = not_allowed



import pytest

from pytest_pl_grader.fixture import FeedbackFixture
from pytest_pl_grader.fixture import StudentFixture


@pytest.mark.grading_data(name="test_exception", points=2)
def test_core_exception(sandbox: StudentFixture, feedback: FeedbackFixture) -> None:
    feedback.set_score(0.1)
    assert sandbox.query("x") == 1



import numpy.linalg as la

index = -1
p_norm = float("inf")

for i, r in enumerate(stock_returns):
    pp = la.norm(r - market_trend, ord=p) / la.norm(r, ord=p)
    if pp < p_norm:
        index = i
        p_norm = la.norm(r, ord=p)



import pytest

from pytest_pl_grader.fixture import DataFixture
from pytest_pl_grader.fixture import FeedbackFixture
from pytest_pl_grader.fixture import StudentFixture

student_code_pattern = "func_name_code*.py"


@pytest.mark.grading_data(name="Test expected inputs", points=4)
def test_input(sandbox: StudentFixture, data_json: DataFixture, feedback: FeedbackFixture) -> None:
    function_name = data_json["params"]["function_name"]
    # student_function = getattr(self.st, function_name)

    num_correct = 0
    pairs_dicts = data_json["params"]["pairs"]

    for pair_dict in pairs_dicts:
        func_input = pair_dict["input"]
        expected_output = pair_dict["output"]
        student_output = sandbox.query_function(function_name, func_input)

        if student_output == expected_output:
            feedback.add_message(f'Function "{function_name}" returned "{student_output}" on input "{func_input}".')
            num_correct += 1

        else:
            feedback.add_message(
                f'Function "{function_name}" returned "{student_output}" on input "{func_input}", not "{expected_output}".'
            )

    percentage_score = num_correct / len(pairs_dicts)
    feedback.set_score(percentage_score)


@pytest.mark.grading_data(name="Test default output", points=1)
def test_default_output(sandbox: StudentFixture, data_json: DataFixture, feedback: FeedbackFixture) -> None:
    function_name = data_json["params"]["function_name"]
    invalid_input = data_json["params"]["invalid_input"]
    default_output = data_json["params"]["default_output"]

    student_output = sandbox.query_function(function_name, invalid_input)

    if student_output == default_output:
        feedback.add_message(f'Function "{function_name}" returned "{student_output}" on input "{invalid_input}".')
        feedback.set_score(1)

    else:
        feedback.add_message(f'Function "{function_name}" returned "{student_output}" on input "{invalid_input}", not "{default_output}".')
        feedback.set_score(0)



import json
import logging
import os
import sys
from collections.abc import Iterable
from copy import deepcopy
from pathlib import Path
from types import ModuleType
from typing import Any
from typing import cast

import _pytest
import _pytest.reports
import _pytest.terminal
import pytest
from _pytest.config import Config
from prettytable import PrettyTable

from .fixture import FeedbackFixture
from .fixture import StudentFiles
from .fixture import StudentFixture
from .utils import GradingOutputLevel
from .utils import NamesForUserInfo
from .utils import ProcessStatusCode
from .utils import get_output_level_marker

logger = logging.getLogger(__name__)


def get_datadir(test_module: ModuleType) -> Path | None:
    """
    Get the data directory for the current test module.
    """

    if test_module is None:
        # In case the test is not in a module (e.g., it is a class method)
        # or a standalone function, you can skip this step
        return None

    # Access the __file__ attribute of the module
    module_filepath_str = test_module.__file__

    if module_filepath_str is None:
        return None

    # Convert it to a pathlib.Path object for easier manipulation
    module_path = Path(module_filepath_str)

    # Let's assume you have a 'data' directory next to your test file
    data_dir = module_path.parent / module_path.stem

    return data_dir


@pytest.fixture(scope="module")
def data_json(request: pytest.FixtureRequest) -> dict[str, Any] | None:
    try:
        datadir = get_datadir(request.module)
        assert datadir is not None
        data_file = datadir / "data.json"
        return json.loads(data_file.read_text(encoding="utf-8"))
    except Exception:
        pass  # TODO add logging

    # If the data file is not found or cannot be read, return None
    return None

    # if datadir is None or not datadir.is_dir():
    #     raise ValueError(f"Data directory '{datadir}' not found or is not a directory.")

    # if not data_file.is_file():
    #     raise ValueError(f"Data file '{data_file.name}' not found in '{datadir}'.")


# TODO maybe change to the module scope??
@pytest.fixture
def sandbox(request: pytest.FixtureRequest, data_json: dict[str, Any] | None) -> Iterable[StudentFixture]:
    # Default timeout TODO make this a command line option?
    initialization_timeout = 1

    if data_json is None:
        params_dict = {}
    else:
        params_dict = data_json.get("params", {})

    import_whitelist = params_dict.get("import_whitelist")
    import_blacklist = params_dict.get("import_blacklist")

    # TODO make sure this contains only valid builtins
    builtin_whitelist = params_dict.get("builtin_whitelist")

    names_for_user_list = cast(list[NamesForUserInfo] | None, params_dict.get("names_for_user"))

    # TODO maybe make it possible to add custom generators for starting variables?
    starting_vars: dict[str, Any] = {
        "__data_params": deepcopy(params_dict) if data_json is not None else {},
    }

    if names_for_user_list is not None:
        for names_dict in names_for_user_list:
            name = names_dict["name"]
            value = params_dict.get(name, None)

            variable_type = type(value).__name__.strip()
            expected_variable_type = names_dict["type"].strip()

            if variable_type != expected_variable_type and value is not None:
                logger.warning(f"Variable type mismatch for starting var {name}: expected {expected_variable_type}, got {variable_type}")

            starting_vars[name] = value

    # Check for the custom mark
    marker = request.node.get_closest_marker("sandbox_timeout")
    if marker:
        # Markers can have positional arguments (args) or keyword arguments (kwargs)
        # We'll assume the timeout is the first positional argument
        if marker.args:
            initialization_timeout = marker.args[0]

    fixture = StudentFixture(
        file_names=request.param,
        import_whitelist=import_whitelist,
        import_blacklist=import_blacklist,
        starting_vars=starting_vars,
        builtin_whitelist=builtin_whitelist,
        names_for_user_list=names_for_user_list,
        worker_username=request.config.getoption("--worker-username"),
    )

    try:
        response = fixture.start_student_code_server(initialization_timeout=initialization_timeout)
        response_status = response["status"]

        if response_status == ProcessStatusCode.EXCEPTION:
            output_level: GradingOutputLevel = get_output_level_marker(request.node.get_closest_marker("output"))

            logger.debug(f"Grading output level set to: {output_level}")
            exception_name = response.get("execution_error", "Unknown error")
            fail_message = f"Student code execution failed with an exception: {exception_name}"

            if output_level == GradingOutputLevel.ExceptionName:
                pytest.fail(fail_message, pytrace=False)

            exception_message = response.get("execution_message", "")
            fail_message += f"{os.linesep}Exception Message: {exception_message}"

            if output_level == GradingOutputLevel.ExceptionMessage:
                pytest.fail(fail_message, pytrace=False)

            # TODO make this not an assert?
            assert output_level == GradingOutputLevel.FullTraceback

            if exception_traceback := response.get("execution_traceback", ""):
                fail_message += f"{os.linesep * 2}{exception_traceback}"

            pytest.fail(fail_message, pytrace=False)

        elif response_status == ProcessStatusCode.TIMEOUT:
            # Don't get the exception message since there usually isn't one for timeouts
            pytest.fail("Student code initialization timed out", pytrace=False)

        elif response_status == ProcessStatusCode.NO_RESPONSE:
            # Don't get the exception message since there usually isn't one for timeouts
            pytest.fail(f"No response from initialization with timeout {initialization_timeout}", pytrace=False)

        elif response_status != ProcessStatusCode.SUCCESS:
            logger.warning(f"Unexpected status in response from student code server: {response}")
            pytest.fail(f"Unexpected status from student code server: {response_status}", pytrace=False)

        yield fixture
    finally:
        fixture._cleanup()


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    """
    TODO this is where the parameterization inside the folder is happening
    """

    # # Let's assume you have a 'data' directory next to your test file
    data_dir = get_datadir(metafunc.module)

    if data_dir is None:
        raise ValueError

    if "sandbox" in metafunc.fixturenames:
        if data_dir.is_dir():
            student_code_pattern = metafunc.module.__dict__.get("student_code_pattern", "student_code*.py")

            # print("IN THE DATA DIR")
            # Find a specific data file, e.g., 'test_data.txt'
            leading_file = data_dir / "leading_code.py"
            trailing_file = data_dir / "trailing_code.py"
            setup_code_file = data_dir / "setup_code.py"

            student_code_files = list(data_dir.glob(student_code_pattern))

            file_tups = [
                StudentFiles(leading_file, trailing_file, student_code_file, setup_code_file) for student_code_file in student_code_files
            ]
            file_stems = [file_tup.student_code_file.stem for file_tup in file_tups]

            metafunc.parametrize("sandbox", file_tups, indirect=True, ids=file_stems)
            # else:
            #    pass
            # pytest.skip(f"Data file '{data_file.name}' not found in '{data_dir}'")
        else:
            pass
            # pytest.skip(f"Data directory '{data_dir}' not found.")


def pytest_addoption(parser: pytest.Parser) -> None:
    """
    Registers command-line options for the plugin.
    """
    group = parser.getgroup("pytest_pl_autograder", "Prairielearn Python Autograder Options")

    group.addoption(
        "--worker-username",
        action="store",
        default=None,
        help="The username for the user of the worker process.",
    )


def _win32_longpath(path):
    """
    Helper function to add the long path prefix for Windows, so that shutil.copytree
     won't fail while working with paths with 255+ chars.
    TODO move this to the utils module.
    From https://github.com/gabrielcnr/pytest-datadir/blob/master/src/pytest_datadir/plugin.py
    """
    if sys.platform == "win32":
        # The use of os.path.normpath here is necessary since "the "\\?\" prefix
        # to a path string tells the Windows APIs to disable all string parsing
        # and to send the string that follows it straight to the file system".
        # (See https://docs.microsoft.com/pt-br/windows/desktop/FileIO/naming-a-file)
        normalized = os.path.normpath(path)
        if not normalized.startswith("\\\\?\\"):
            is_unc = normalized.startswith("\\\\")
            # see https://en.wikipedia.org/wiki/Path_(computing)#Universal_Naming_Convention
            if is_unc:  # then we need to insert an additional "UNC\" to the longpath prefix
                normalized = normalized.replace("\\\\", "\\\\?\\UNC\\")
            else:
                normalized = "\\\\?\\" + normalized
        return normalized
    else:
        return path


def pytest_runtest_setup(item):
    # TODO clean up this function
    marker = item.get_closest_marker("sandbox")
    if marker:
        if marker.args:
            raise ValueError("benchmark mark can't have positional arguments.")
        for name in marker.kwargs:
            if name not in (
                "max_time",
                "min_rounds",
                "min_time",
                "timer",
                "group",
                "disable_gc",
                "warmup",
                "warmup_iterations",
                "calibration_precision",
                "cprofile",
            ):
                raise ValueError(f"benchmark mark can't have {name!r} keyword argument.")


@pytest.hookimpl(trylast=True)  # force the other plugins to initialise, fixes issue with capture not being properly initialized
def pytest_configure(config: Config) -> None:
    # config.addinivalue_line("markers", "benchmark: mark a test with custom benchmark settings.")
    # bs = config._benchmarksession = BenchmarkSession(config)
    # bs.handle_loading()
    # config.pluginmanager.register(bs, "pytest-benchmark")

    # Add a marker for the sandbox fixture to set the initialization timeout
    config.addinivalue_line("markers", "sandbox_timeout(timeout_value): sets a timeout for initialization of the sandbox fixture")

    # Only register our plugin if it hasn't been already (e.g., in case of multiple conftests)
    if not hasattr(config, "result_collector_plugin"):
        config.result_collector_plugin = ResultCollectorPlugin()  # type: ignore[attr-defined]
        config.pluginmanager.register(config.result_collector_plugin)  # type: ignore[attr-defined]


class ResultCollectorPlugin:
    collected_results: dict[str, tuple[_pytest.reports.TestReport, pytest.CallInfo]]
    student_feedback_data: dict[str, FeedbackFixture]
    grading_data: dict[str, Any]

    def __init__(self) -> None:
        self.collected_results = {}
        self.student_feedback_data = {}
        self.grading_data = {}

    def pytest_configure(self, config: Config) -> None:
        """
        Register our custom marker to avoid warnings.
        """
        config.addinivalue_line("markers", "grading_data(name, points): Mark a test with custom data that can be injected.")

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_makereport(self, item: pytest.Item, call: pytest.CallInfo) -> Iterable[None]:
        """
        Hook wrapper to capture test outcomes.
        """
        outcome = yield
        report: _pytest.reports.TestReport = outcome.get_result()  # type: ignore[attr-defined]
        marker = item.get_closest_marker("grading_data")  # Ensure the marker is registered

        if marker:
            self.grading_data[item.nodeid] = marker.kwargs

        # Make a report for the setup phase, replace with the call phase if it happens later
        if report.when == "setup":
            self.collected_results[report.nodeid] = (report, call)
            # Add a default outcome if not already set

        elif report.when == "call":
            self.collected_results[report.nodeid] = (report, call)
            # You could store more details here if needed
            # item.config.my_test_results[report.nodeid] = {
            #     "outcome": report.outcome,
            #     "duration": report.duration,
            # }

        fixture = None
        if hasattr(item, "funcargs"):
            student_code_fixture = item.funcargs.get("sandbox")
            feedback_fixture = item.funcargs.get("feedback")

        if fixture is not None and not isinstance(fixture, StudentFixture):
            pass
            # raise TypeError(
            #     f"unexpected type for `benchmark` in funcargs, {fixture!r} must be a BenchmarkFixture instance. "
            #     "You should not use other plugins that define a `benchmark` fixture, or return and unexpected value if you do redefine it."
            # )
        # if fixture:
        #     fixture.skipped = outcome.get_result().outcome == "skipped"

    @pytest.fixture
    def feedback(self, request: pytest.FixtureRequest) -> FeedbackFixture:
        """
        A fixture that allows tests to add feedback messages and scores.
        """
        nodeid = request.node.nodeid

        # Initialize feedback for this test if it doesn't exist
        if nodeid not in self.student_feedback_data:
            self.student_feedback_data[nodeid] = FeedbackFixture(test_id=nodeid)

        return self.student_feedback_data[nodeid]

    @pytest.hookimpl(hookwrapper=True)
    def pytest_sessionfinish(self, session: pytest.Session, exitstatus: int) -> Iterable[None]:
        """
        Hook wrapper to process test results after the session finishes.
        """
        yield  # Let other sessionfinish hooks run

        # print("\n--- Custom Test Results Summary (via Plugin Class) ---")
        # for nodeid, outcome in self.collected_results.items():
        #     print(f"Test: {nodeid} -> Outcome: {outcome}")
        # print("--------------------------------------------------")

        # # Example: Check the result of a specific test by its nodeid
        # target_nodeid = "test_example.py::test_passing_example" # Replace with a test you have
        # if target_nodeid in self.collected_results:
        #     print(f"\nResult for '{target_nodeid}': {self.collected_results[target_nodeid]}")
        # else:
        #     print(f"\n'{target_nodeid}' not found or no results collected.")

        # Collect all student feedback and generate the final report.
        final_results = []

        for item in session.items:
            nodeid = item.nodeid

            # for nodeid, feedback_obj in self.student_feedback_data.items():
            grading_data = self.grading_data.setdefault(nodeid, {"name": nodeid, "points": 1})

            if nodeid not in self.collected_results:
                continue  # Skip if no results collected for this test

            report, call = self.collected_results[nodeid]
            outcome = report.outcome

            if nodeid in self.student_feedback_data:
                feedback_obj = self.student_feedback_data[nodeid]
            else:
                # Create an empty feedback object if none was created during the test
                feedback_obj = FeedbackFixture(test_id=nodeid)

            # If the test failed, we can add the exception message to the feedback
            if report.outcome == "failed" and call.excinfo is not None:
                output_level: GradingOutputLevel = get_output_level_marker(item.get_closest_marker("output"))

                logger.debug(f"Grading output level set to: {output_level}")

                if output_level == GradingOutputLevel.ExceptionName:
                    exception_name = call.excinfo.type.__name__
                    fail_message = f"Student code grading failed with an exception: {exception_name}"
                    feedback_obj.add_message(fail_message)

                elif output_level == GradingOutputLevel.ExceptionMessage:
                    exception_name = call.excinfo.type.__name__
                    # TODO make this work with multiline messages somehow?
                    exception_message = str(call.excinfo.value).split(os.linesep)[0]
                    fail_message = (
                        f"Student code grading failed with an exception: {exception_name}{os.linesep}Exception Message: {exception_message}"
                    )
                    feedback_obj.add_message(fail_message)

                # If showing more than the exception name, show the message + full traceback
                else:
                    feedback_obj.add_message(str(call.excinfo.getrepr(style="no")))

            res_obj = feedback_obj.to_dict()
            res_obj["name"] = grading_data.get("name", nodeid)
            res_obj["max_points"] = grading_data.get("points", 1)

            if report.when in ["setup", "teardown"] and report.outcome == "failed":
                res_obj["outcome"] = "error"
            else:
                res_obj["outcome"] = outcome

            if outcome == "passed":
                if not feedback_obj.final_score_override:
                    res_obj["points_frac"] = 1.0
                # Otherwise, we just use the set points value

            elif res_obj["points_frac"] is None:
                if outcome == "failed":
                    res_obj["points_frac"] = 0.0
                else:
                    # TODO fill in logic for other outcomes
                    # e.g., "skipped", "xpassed", etc.
                    # For now, we raise an error for unexpected outcomes
                    raise ValueError(f"Unexpected outcome '{outcome}' for test '{nodeid}'.")

            res_obj["points"] = res_obj["points_frac"] * res_obj["max_points"]
            final_results.append(res_obj)
        # TODO add gradable property
        # https://prairielearn.readthedocs.io/en/latest/externalGrading/#grading-results

        total_score = sum(res["points_frac"] * res["max_points"] for res in final_results)

        # TODO should probably just raise an exception if this is zero bc it's almost certainly a mistake
        total_possible_score = sum(res["max_points"] for res in final_results)

        res_dict = {
            "score": total_score / total_possible_score if total_possible_score > 0 else 0,
            # TODO figure out something useful to put here (maybe single source of compilation failure??)
            # "output": "Overall feedback for the autograder session.",
            "tests": final_results,
        }

        print_autograder_summary(session, final_results)

        # Example: Save to a JSON file
        # TODO make this configurable via command line options
        output_path = session.config.rootpath / "autograder_results.json"
        with open(output_path, "w") as f:
            json.dump(res_dict, f, indent=4, sort_keys=True)
        print(f"\nAutograder results saved to {output_path}")

        # For autograding platforms like Gradescope, you might format
        # it according to their specific JSON format.
        # Example Gradescope format:
        # {
        #   "score": 0,
        #   "output": "Overall feedback",
        #   "tests": [
        #     {"name": "Test Case 1", "score": 2, "max_score": 5, "output": "Feedback for test 1"},
        #     ...
        #   ]
        # }


def print_autograder_summary(session: pytest.Session, test_results: list[dict[str, Any]]) -> None:
    """
    Print a summary of the autograder results in a formatted table.
    This function is called at the end of the test session to display
    the results in a readable format.
    """
    if not session.config.pluginmanager.hasplugin("terminalreporter"):
        print("Terminal reporter plugin not found. Cannot print autograder summary.")
        return

    # Get the terminal reporter and its writer
    reporter: _pytest.terminal.TerminalReporter = session.config.pluginmanager.getplugin("terminalreporter")
    writer = reporter._tw  # Access the internal TerminalWriter instance

    if not test_results:
        writer.line("No tests were run or no results collected.")
        return

    # Create a PrettyTable instance
    table = PrettyTable()
    table_headers = ["Test Name", "Score", "Feedback"]
    table.field_names = table_headers

    # Set alignment for columns
    table.align["Test Name"] = "l"
    table.align["Score"] = "c"
    table.align["Feedback"] = "l"

    # Add data rows
    for result in test_results:
        table.add_row([result["test_id"], result["points"], result["message"]])

    # TODO this is incorrect as the fractions were changed in the display shown to PL.
    # Can fix this later since most people will just see the output shown in PL.

    # Calculate total score
    total_score = sum(result["points"] for result in test_results)
    max_score = len(test_results)

    # Add total score row
    table.add_row(["Total Score", f"{total_score}/{max_score}", ""])

    # Set table style (optional, but 'grid' is similar to previous tabulate output)
    # You can experiment with other styles like:
    # table.set_style(MARKDOWN)
    # table.set_style(SINGLE_BORDER)
    # table.set_style(DOUBLE_BORDER)
    # For a grid-like appearance, PrettyTable's default is quite good,
    # or you can explicitly set it to something like:
    # from prettytable import MSWORD_FRIENDLY, PLAIN_COLUMNS, ORGMODE
    # table.set_style(MSWORD_FRIENDLY)

    # Try to make the table fit the terminal width
    # PrettyTable's get_string method has a 'max_width' parameter for columns
    # We distribute the available width among the columns.
    terminal_width = writer.fullwidth

    # Estimate average column width, subtracting for borders and padding
    # This is a heuristic, PrettyTable will adjust
    num_columns = len(table_headers)
    estimated_col_width = (terminal_width // num_columns) - 4  # Subtract for borders/padding

    # Set max_width for each column to attempt fitting
    # PrettyTable will wrap text if it exceeds max_width
    for field in table.field_names:
        table.max_width[field] = estimated_col_width

    # Generate the table string
    # prettytable's get_string() will automatically handle column width adjustments
    # and wrapping based on max_width and terminal size.
    table_string = table.get_string()

    # Get the width of the generated table from the first line
    # Prettytable's output includes the borders, so this should be accurate.
    table_width = len(table_string.splitlines()[0])

    # Print the autograder summary as a centered header above the table
    summary_title = "Autograder Summary"
    # Calculate padding for centering. Subtract 2 for the outer '+' characters.
    # If the title is longer than the table width, just print it without centering.
    if table_width > len(summary_title) + 2:
        padding_left = (table_width - len(summary_title) - 2) // 2
        padding_right = (table_width - len(summary_title) - 2) - padding_left
        header_line = f"+{'=' * padding_left} {summary_title} {'=' * padding_right}+"
    else:
        # Fallback if the table is very narrow
        header_line = f"+{summary_title.center(table_width - 2)}+"

    writer.line(os.linesep)  # Add a newline before the custom header
    writer.line(header_line, bold=True)

    # Print the generated table using the TerminalWriter
    for line in table_string.splitlines():
        writer.line(line)

    writer.write(f"{os.linesep}Final Grade: {total_score}/{max_score}{os.linesep}", bold=True)



import numpy as np
import numpy.linalg as la


def solve_ratings(members, sale_metadata):
    sale_metadata = [
        {"hours": [("Bob", 50), ("Greg", 45), ("Susan", 25)], "units_sold": 160},
        {"hours": [("Greg", 50), ("Bob", 25), ("Susan", 70)], "units_sold": 180},
        {"hours": [("Bob", 35), ("Susan", 45), ("Greg", 25)], "units_sold": 150},
    ]

    members = ["Bob", "Susan", "Greg"]
    sales = np.zeros(len(sale_metadata))
    theMatrix = np.zeros((len(members), len(sale_metadata)))
    for i in range(len(sale_metadata)):
        for j in range(len(sale_metadata[i]["hours"])):
            col = members.index(sale_metadata[i]["hours"][j][0])
            theMatrix[i][col] = sale_metadata[i]["hours"][j][1]
        sales[i] = sale_metadata[i]["units_sold"]
    ratings = la.solve(theMatrix, sales)
    highest_performing = np.max(ratings)
    lowest_performing = np.min(ratings)
    member_ratings = dict(zip(members, ratings))

    return member_ratings, highest_performing, lowest_performing



import math

import numpy as np
import pytest

from pytest_pl_grader.fixture import DataFixture
from pytest_pl_grader.fixture import FeedbackFixture
from pytest_pl_grader.fixture import StudentFixture


def factorial(n):
    result = 1
    for i in range(1, n + 1):
        result *= i
    return result


def f(x, sin_coefficient, cos_coefficient, constant_term):
    return np.sin(sin_coefficient * x) + cos_coefficient * np.cos(x) + constant_term


def taylor_f(x, sin_coefficient, cos_coefficient, constant_term, taylor_degree):
    t0 = constant_term + cos_coefficient
    t1 = sin_coefficient
    t2 = -cos_coefficient
    t3 = -(sin_coefficient**3)
    taylor_sum = t0 + t1 * x + (t2 / factorial(2)) * (x**2) + (t3 / factorial(3)) * (x**3)
    if taylor_degree == 4:
        t4 = cos_coefficient
        taylor_sum += (t4 / factorial(4)) * (x**4)
    return taylor_sum


def analyze_vibration(x_hat, threshold, sin_coefficient, cos_coefficient, constant_term, taylor_degree):
    approximation = taylor_f(x_hat, sin_coefficient, cos_coefficient, constant_term, taylor_degree)
    true_value = f(x_hat, sin_coefficient, cos_coefficient, constant_term)
    production_ready = approximation > threshold
    return approximation, true_value, production_ready


@pytest.mark.grading_data(name="Test analyze_vibration function", points=6)
@pytest.mark.sandbox_timeout(5)
@pytest.mark.parametrize(
    ("i", "x_hat", "threshold"),
    [(1, 0.7235132915688036, 23.137774021447807), (2, 0.23796317102834275, 23.50476117422776)],
)
def test_0(sandbox: StudentFixture, feedback: FeedbackFixture, data_json: DataFixture, i: int, x_hat: float, threshold: float) -> None:
    points = 0

    sin_coefficient = data_json["params"]["sin_coefficient"]
    cos_coefficient = data_json["params"]["cos_coefficient"]
    constant_term = data_json["params"]["constant_term"]
    taylor_degree = data_json["params"]["taylor_degree"]

    feedback.add_message(" ")
    feedback.add_message(f"Test #{i}")

    st_result = sandbox.query_function("analyze_vibration", x_hat, threshold)
    ref_result = analyze_vibration(
        x_hat,
        threshold,
        sin_coefficient,
        cos_coefficient,
        constant_term,
        taylor_degree,
    )

    assert st_result is not None, "Your function did not return any value. Ensure you have a return statement."

    assert isinstance(st_result, list) and len(st_result) == 3, "Your function should return a tuple with exactly three elements."

    points = 0

    if math.isclose(ref_result[0], st_result[0]):
        points += 1

    if math.isclose(ref_result[1], st_result[1]):
        points += 1

    if ref_result[2] == st_result[2]:
        points += 1
        feedback.add_message("'production_ready' looks good")
    else:
        feedback.add_message(f"'production_ready' is incorrect and has type {type(st_result[2])}")

    feedback.set_score_final(points / 3)



import math

largest_fp = 2**7 * (1 - 2 ** (-4))
machine_epsilon = 2 ** (-3)

start_exp = math.floor(math.log(start, 2))
end_exp = math.floor(math.log(end, 2))

start_frac = start / 2**start_exp
end_frac = end / 2**end_exp

fp_lst = []

for e in range(start_exp, end_exp + 1):
    for i in range(8):
        if e == start_exp and 1 + float(i) / 8 < start_frac:
            continue
        if e == end_exp and 1 + float(i) / 8 > end_frac:
            continue
        fp_lst.append((1 + float(i) / 8) * 2**e)



from typing import TypedDict

import numpy as np
import numpy.linalg as la
import pytest

from pytest_pl_grader.fixture import FeedbackFixture
from pytest_pl_grader.fixture import StudentFixture


class ParameterDict(TypedDict):
    hours: list[tuple[str, np.int_]]
    units_sold: np.int_


def compute_starting_parameters() -> tuple[list[str], list[ParameterDict]]:
    overall_members = ["Alex", "Jordan", "Taylor", "Casey", "Riley", "Chris", "Avery", "Sam", "Devon", "Morgan"]

    rng = np.random.default_rng()

    members = list(rng.choice(overall_members, size=5, replace=False))
    N_sales = len(members)

    hours_set = set()
    sale_metadata: list[ParameterDict] = []

    # Avoid edge case where two lists of hours are the same, causing a singularity
    while len(sale_metadata) < N_sales:
        hours = tuple(rng.integers(10, 101, size=N_sales))
        if hours in hours_set:
            continue
        hours_set.add(hours)
        rng.shuffle(members)
        sale_metadata.append({"hours": list(zip(members, hours)), "units_sold": rng.integers(100, 2001)})

    return members, sale_metadata


def compute_answers(members: list[str], sale_metadata: list[ParameterDict]) -> tuple[dict[str, float], str, str]:
    member_id = {member: i for i, member in enumerate(members)}
    A = np.zeros((len(sale_metadata), len(members)))
    sales = np.zeros(len(sale_metadata))

    for i, sale_md in enumerate(sale_metadata):
        for v, t in sale_md["hours"]:
            A[i, member_id[v]] = t
        sales[i] = sale_md["units_sold"]

    contribs = la.solve(A, sales)
    member_ratings = {members[i]: contribs[i] for i in range(len(members))}
    highest_performing = members[np.argmax(contribs)]
    lowest_performing = members[np.argmin(contribs)]

    return member_ratings, highest_performing, lowest_performing


@pytest.mark.grading_data(name="ratings", points=4)
def test_0(sandbox: StudentFixture, feedback: FeedbackFixture) -> None:
    score = 0
    members, sale_metadata = compute_starting_parameters()

    student_member_ratings, _, _ = sandbox.query_function("solve_ratings", members, sale_metadata)

    ref_member_ratings, _, _ = compute_answers(members, sale_metadata)

    if student_member_ratings is not None:
        if type(student_member_ratings) != dict:
            feedback.add_message("member_ratings is not a dictionary")
        else:
            for member, rating in student_member_ratings.items():
                if member not in ref_member_ratings:
                    feedback.add_message(f"{member} is not a known team member")
                    continue
                if ref_member_ratings[member] != rating:
                    feedback.add_message(f"{member} has an incorrect rating")
                else:
                    score += 1
    else:
        feedback.add_message("member_ratings is not defined")

    feedback.set_score_final(score / len(ref_member_ratings))


@pytest.mark.grading_data(name="highest_performing", points=1)
def test_1(sandbox: StudentFixture) -> None:
    members, sale_metadata = compute_starting_parameters()

    _, student_highest_performing, _ = sandbox.query_function("solve_ratings", members, sale_metadata)

    _, ref_highest_performing, _ = compute_answers(members, sale_metadata)

    assert ref_highest_performing == student_highest_performing, "highest_performing is not correct"


@pytest.mark.grading_data(name="lowest_performing", points=1)
def test_2(sandbox: StudentFixture) -> None:
    members, sale_metadata = compute_starting_parameters()

    _, _, student_lowest_performing = sandbox.query_function("solve_ratings", members, sale_metadata)

    _, _, ref_lowest_performing = compute_answers(members, sale_metadata)

    assert ref_lowest_performing == student_lowest_performing, "lowest_performing is not correct"



import numpy as np


def analyze_vibration(x_hat, threshold):
    approximation = 1
    true_value = np.sin(6 * x_hat) + 3 * np.cos(x_hat) + 6
    production_ready = 9

    return approximation, true_value, production_ready



import json
import logging
import os
import socket
import subprocess
import sys
from importlib.resources import files
from pathlib import Path
from typing import Any
from typing import NamedTuple

from .json_utils import from_json
from .utils import NamesForUserInfo
from .utils import ProcessStartRequest
from .utils import ProcessStartResponse
from .utils import ProcessStatusCode
from .utils import SetupQueryRequest
from .utils import SetupQueryResponse
from .utils import StudentFunctionRequest
from .utils import StudentFunctionResponse
from .utils import StudentQueryRequest
from .utils import StudentQueryResponse
from .utils import deserialize_object_unsafe
from .utils import drop_privileges
from .utils import serialize_object_unsafe

DataFixture = dict[str, Any]

SCRIPT_PATH = str(files("pytest_pl_grader").joinpath("_student_code_runner.py"))
DEFAULT_TIMEOUT = 1.0

logger = logging.getLogger(__name__)


class StudentFiles(NamedTuple):
    leading_file: Path
    trailing_file: Path
    student_code_file: Path
    setup_code_file: Path


class FeedbackFixture:
    """
    A fixture to handle feedback from the student code.
    """

    test_id: str
    messages: list[str]
    score: float | None
    final_score_override: bool

    def __init__(self, test_id: str) -> None:
        self.test_id = test_id
        self.messages = []
        self.score = None
        self.final_score_override = False

    def add_message(self, message: str) -> None:
        self.messages.append(message)

    def set_score(self, score: float) -> None:
        self.score = score

    def set_score_final(self, score: float) -> None:
        """
        Sets the final score for the test. This should be called at the end of the test.
        """
        if self.score is not None:
            raise RuntimeError("Final score has already been set.")

        # TODO maybe change this to assert the score is 1? Then it will fail if the score is not 1.
        # Will maintain invariant that score should be 1 if all tests pass.
        self.score = score
        self.final_score_override = True

    def to_dict(self) -> dict:
        return {
            "test_id": self.test_id,
            "message": os.linesep.join(self.messages),
            "points_frac": self.score,
        }


class StudentFixture:
    process: subprocess.Popen | None
    leading_file: Path
    trailing_file: Path
    student_code_file: Path
    student_socket: socket.socket | None
    import_whitelist: list[str] | None
    import_blacklist: list[str] | None
    starting_vars: dict[str, Any] | None
    builtin_whitelist: list[str] | None
    names_for_user_list: list[NamesForUserInfo] | None
    worker_username: str | None

    def __init__(
        self,
        file_names: StudentFiles,
        import_whitelist: list[str] | None,
        import_blacklist: list[str] | None,
        starting_vars: dict[str, Any] | None,
        builtin_whitelist: list[str] | None,
        names_for_user_list: list[NamesForUserInfo] | None,
        worker_username: str | None,
    ) -> None:
        self.leading_file = file_names.leading_file
        self.trailing_file = file_names.trailing_file
        self.student_code_file = file_names.student_code_file
        self.setup_code_file = file_names.setup_code_file

        self.import_whitelist = import_whitelist
        self.import_blacklist = import_blacklist
        self.starting_vars = starting_vars
        self.builtin_whitelist = builtin_whitelist
        self.names_for_user_list = names_for_user_list
        self.worker_username = worker_username

        # Initialize the process and socket to None
        self.process = None
        self.student_socket = None

    def _assert_process_running(self) -> None:
        """
        TODO make the type of this a typeguard for process and socket
        """

        assert self.process is not None, "Student code server process is not running. Please start it first."

        process_return_code = self.process.poll()
        if process_return_code is not None:
            raise RuntimeError(f"Student code server process terminated with code {process_return_code}.")

    def _send_json_object(
        self, json_object: StudentQueryRequest | ProcessStartRequest | StudentFunctionRequest | SetupQueryRequest
    ) -> None:
        """
        Sends a JSON object to the student code server.
        """
        assert self.student_socket is not None, "Student socket is not connected. Please start the student code server first."
        self.student_socket.sendall((json.dumps(json_object) + os.linesep).encode("utf-8"))

    def _read_from_socket(self) -> bytes:
        """
        Reads data from a socket until a termination character is found.
        """
        buffer = bytearray()

        terminator = os.linesep.encode("utf-8")
        max_len: int | None = None  # TODO add max length parameter?

        assert self.student_socket is not None, "Student socket is not connected. Please start the student code server first."

        # Define a small chunk size for reading
        chunk_size = 4096
        chunk: bytes = b""

        # TODO maybe set a hard iteration limit to avoid infinite loops?
        while (idx := chunk.rfind(terminator)) == -1:
            try:
                # Read a chunk of data
                chunk = self.student_socket.recv(chunk_size)
            except TimeoutError as e:
                # Re-raise the timeout error
                raise TimeoutError("Socket read timed out.") from e

            # Check if the connection was closed
            if not chunk:
                # Connection closed before terminator was found
                raise Exception("Connection closed by peer before termination character was found.")

            # Append the new chunk to the buffer
            buffer.extend(chunk)

            # Check for maximum length constraint
            if max_len is not None and len(buffer) >= max_len:
                # Raise an error or return buffer depending on desired behavior
                raise Exception(f"Maximum read length of {max_len} exceeded.")

        loc = len(buffer) - len(chunk) + idx + len(terminator)

        # Return the buffer content up to and including the terminator
        # to only return the necessary data if more data was read
        return buffer[:loc]

    def start_student_code_server(self, *, initialization_timeout: float = DEFAULT_TIMEOUT) -> ProcessStartResponse:
        if self.worker_username is not None:
            logger.debug(f"Starting student code server with worker username: {self.worker_username}")
        else:
            logger.debug("Starting student code server without dropping privileges.")

        def try_drop_privileges() -> None:
            if self.worker_username is not None:
                drop_privileges(self.worker_username)

        self.process = subprocess.Popen(
            args=(sys.executable, SCRIPT_PATH),
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            preexec_fn=try_drop_privileges,
        )

        # Assert process is running after popen call
        self._assert_process_running()

        student_code = ""
        if self.leading_file.is_file():
            student_code += self.leading_file.read_text(encoding="utf-8")
            student_code += os.linesep

        if self.student_code_file.is_file():
            student_code += self.student_code_file.read_text(encoding="utf-8")

        if self.trailing_file.is_file():
            student_code += os.linesep
            student_code += self.trailing_file.read_text(encoding="utf-8")

        # TODO maybe add an error message for this?
        setup_code = None
        if self.setup_code_file.is_file():
            setup_code = self.setup_code_file.read_text(encoding="utf-8")

        # TODO make this a shared type
        json_message: ProcessStartRequest = {
            "message_type": "start",
            "student_code": student_code,
            "student_file_name": str(self.student_code_file),
            "setup_code": setup_code,
            "initialization_timeout": initialization_timeout,
            "import_whitelist": self.import_whitelist,
            "import_blacklist": self.import_blacklist,
            "starting_vars": self.starting_vars,
            "builtin_whitelist": self.builtin_whitelist,
            "names_for_user_list": self.names_for_user_list,
        }

        assert self.process.stdout is not None, "Process stdout is None. Ensure the process is started correctly."

        line = self.process.stdout.readline().decode()  # Read the initial output from the process to ensure it's ready
        host, port = line.strip().split(",")

        self.student_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.student_socket.settimeout(initialization_timeout)
        self.student_socket.connect((host, int(port)))

        self._send_json_object(json_message)

        try:
            data = self._read_from_socket().decode()
            res: ProcessStartResponse = json.loads(data)
        except Exception as e:
            res = {
                "status": ProcessStatusCode.NO_RESPONSE,
                "execution_error": type(e).__name__,
                "execution_message": str(e),
                "execution_traceback": "",
                "stdout": "",
                "stderr": "",
            }

        return res

    def query_setup_raw(self, var_to_query: str) -> SetupQueryResponse:
        self._assert_process_running()

        json_message: SetupQueryRequest = {"message_type": "query_setup", "var": var_to_query}

        assert self.student_socket is not None, "Student socket is not connected. Please start the student code server first."
        self._send_json_object(json_message)
        data: SetupQueryResponse = json.loads(self._read_from_socket().decode())

        return data

    def query_setup(self, var_to_query: str) -> Any:
        """
        Queries a variable from the setup code and returns its value.
        """
        response = self.query_setup_raw(var_to_query)

        if response["status"] == "not_found":
            raise NameError(f"Query for setup variable '{var_to_query}' failed")

        return deserialize_object_unsafe(response["value_encoded"])

    def query_raw(self, var_to_query: str, *, query_timeout: float = DEFAULT_TIMEOUT) -> StudentQueryResponse:
        self._assert_process_running()

        json_message: StudentQueryRequest = {"message_type": "query", "var": var_to_query, "query_timeout": query_timeout}

        assert self.student_socket is not None, "Student socket is not connected. Please start the student code server first."
        self.student_socket.settimeout(query_timeout)
        self._send_json_object(json_message)
        data: StudentQueryResponse = json.loads(self._read_from_socket().decode())

        return data

    def query(self, var_to_query: str, *, query_timeout: float = DEFAULT_TIMEOUT) -> Any:
        """
        Queries a variable from the student code and returns its value.
        """
        response = self.query_raw(var_to_query, query_timeout=query_timeout)

        if response["status"] == "not_found":
            raise NameError(f"Query for '{var_to_query}' failed")

        return from_json(response["value"])

    def query_function_raw(self, function_name: str, *args, query_timeout: float = DEFAULT_TIMEOUT, **kwargs) -> StudentFunctionResponse:
        """
        TODO add query timeout keyword only argument
        """

        json_message: StudentFunctionRequest = {
            "message_type": "query_function",
            "function_name": function_name,
            "args_encoded": serialize_object_unsafe(args),
            "kwargs_encoded": serialize_object_unsafe(kwargs),
            "query_timeout": query_timeout,
        }

        assert self.student_socket is not None, "Student socket is not connected. Please start the student code server first."
        self.student_socket.settimeout(query_timeout)
        self.student_socket.sendall((json.dumps(json_message) + os.linesep).encode("utf-8"))
        data: StudentFunctionResponse = json.loads(self._read_from_socket().decode())

        return data

    def query_function(self, function_name: str, *args, query_timeout: float = DEFAULT_TIMEOUT, **kwargs) -> Any:
        """
        Queries a function from the student code and returns its return value.
        """
        response = self.query_function_raw(function_name, *args, query_timeout=query_timeout, **kwargs)

        match response["status"]:
            case "exception":
                raise RuntimeError(
                    f"Function '{function_name}' raised an exception {response['exception_name']}: {response['exception_message']}\n{response['traceback']}"
                )
            case "timeout":
                raise TimeoutError(f"Query for function '{function_name}' timed out after {query_timeout} seconds.")
            case "not_found":
                raise NameError(f"Query for function '{function_name}' failed: {response['exception_message']}")

        return from_json(response["value"])

    # TODO add functions that let instructors use the student fixture
    # use the stuff pete set up here: https://github.com/reteps/pytest-autograder-prototype
    def _cleanup(self) -> None:
        if self.student_socket is not None:
            self.student_socket.close()
            self.student_socket = None

        if self.process is not None:
            self.process.terminate()
            self.process.wait()
            self.process = None

    def __repr__(self) -> str:
        return f"StudentFixture(leading_file={self.leading_file}, trailing_file={self.trailing_file}, student_code_file={self.student_code_file})"



import asyncio
import concurrent.futures
import io
import json
import linecache
import os
import sys
import traceback
import types
from collections.abc import Callable
from collections.abc import Mapping
from collections.abc import Sequence
from contextlib import redirect_stderr
from contextlib import redirect_stdout
from copy import deepcopy
from typing import Any

from pytest_pl_grader.json_utils import from_server_json

# TODO make it so that other files in this package cannot import from this one
# ask Gemini how to do it
from pytest_pl_grader.json_utils import to_json
from pytest_pl_grader.utils import FunctionStatusCode
from pytest_pl_grader.utils import NamesForUserInfo
from pytest_pl_grader.utils import ProcessStartRequest
from pytest_pl_grader.utils import ProcessStartResponse
from pytest_pl_grader.utils import ProcessStatusCode
from pytest_pl_grader.utils import QueryStatusCode
from pytest_pl_grader.utils import SetupQueryRequest
from pytest_pl_grader.utils import SetupQueryResponse
from pytest_pl_grader.utils import StudentFunctionRequest
from pytest_pl_grader.utils import StudentFunctionResponse
from pytest_pl_grader.utils import StudentQueryRequest
from pytest_pl_grader.utils import StudentQueryResponse
from pytest_pl_grader.utils import deserialize_object_unsafe
from pytest_pl_grader.utils import get_builtins
from pytest_pl_grader.utils import serialize_object_unsafe

ImportFunction = Callable[[str, Mapping[str, object] | None, Mapping[str, object] | None, Sequence[str], int], types.ModuleType]

HOST = "127.0.0.1"  # Loopback address, means "this computer only"


# Global ThreadPoolExecutor for CPU-bound tasks
# It's good practice to create this once and reuse it.
# The number of workers should ideally be around the number of CPU cores.
executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)


def populate_linecache(contents: str, fname: str) -> None:
    """
    TODO do what's in this file here
    https://github.com/PrairieLearn/PrairieLearn/commit/28c1f0bfb3792c950e5df30061469bfaf0ca199f
    """
    linecache.cache[fname] = (
        len(contents),
        None,
        [line + os.linesep for line in contents.splitlines()],
        fname,
    )


async def student_function_runner(
    student_code_vars: dict[str, Any], func_name: str, timeout: float, args_tup: Any, kwargs_dict: Any
) -> StudentFunctionResponse:
    stdout_capture = io.StringIO()
    stderr_capture = io.StringIO()
    execution_error = None
    exception_traceback = None
    result = None

    try:

        def student_function_temp() -> Any:
            student_function = student_code_vars[func_name]
            return student_function(*args_tup, **kwargs_dict)

        with redirect_stdout(stdout_capture), redirect_stderr(stderr_capture):
            result = await asyncio.wait_for(asyncio.get_event_loop().run_in_executor(executor, student_function_temp), timeout=timeout)
    except Exception as e:
        execution_error = e
        exception_traceback = traceback.format_exc(limit=-1)

    function_response: StudentFunctionResponse = {
        "status": FunctionStatusCode.SUCCESS if execution_error is None else FunctionStatusCode.EXCEPTION,
        "value": to_json(result),
        "stdout": stdout_capture.getvalue(),
        "stderr": stderr_capture.getvalue(),
        "exception_name": type(execution_error).__name__,
        "exception_message": str(execution_error) if execution_error else None,
        "traceback": exception_traceback,
    }

    return function_response


def get_custom_importer(import_whitelist: list[str] | None, import_blacklist: list[str] | None) -> ImportFunction:
    """
    Returns a custom import function that restricts imports based on the provided whitelist and blacklist.
    If a whitelist is provided, only those modules can be imported.
    """

    original_import = __import__

    def custom_import(
        name: str,
        globals: Mapping[str, object] | None = None,
        locals: Mapping[str, object] | None = None,
        fromlist: Sequence[str] = (),
        level: int = 0,
    ) -> types.ModuleType:
        # Allow specific modules to be imported
        if import_blacklist is not None and name in import_blacklist:
            raise ImportError(f"Module '{name}' is blacklisted and cannot be imported.")
        elif (
            (import_whitelist is not None and name in import_whitelist)
            or name.startswith("__")  # Allow internal dunder imports if necessary for basic functionality
            or import_whitelist is None
        ):
            return original_import(name, globals, locals, fromlist, level)
        else:
            # Forbid other imports
            raise ImportError(f"Module '{name}' is not allowed to be imported.")

    return custom_import


async def student_code_runner(
    setup_code: str | None,
    student_code: str,
    student_file_name: str,
    timeout: float,
    import_whitelist: list[str] | None,
    import_blacklist: list[str] | None,
    starting_vars: dict[str, Any] | None,
    builtin_whitelist: list[str] | None,
    names_for_user_list: list[NamesForUserInfo] | None,
) -> tuple[dict[str, Any], dict[str, Any], ProcessStartResponse]:
    stdout_capture = io.StringIO()
    stderr_capture = io.StringIO()
    execution_error = None
    exception_traceback = None
    local_vars = deepcopy(starting_vars) if starting_vars else {}
    local_vars["__from_server_json"] = from_server_json  # Add the deserialization function to the local variables for setup code to use

    student_code_vars: dict[str, Any] = {}
    student_code_vars["__builtins__"] = get_builtins(builtin_whitelist)

    student_code_vars["__builtins__"]["__name__"] = "__main__"  # Set __name__ to "__main__" to mimic the main module
    student_code_vars["__builtins__"]["__import__"] = get_custom_importer(import_whitelist, import_blacklist)

    # TODO the data object is not passed into the setup code. Add this if needed.

    try:
        # First, execute the setup code if provided
        if setup_code:
            # Compile the setup code
            code_setup = compile(setup_code, "<setup>", "exec")
            with redirect_stdout(stdout_capture), redirect_stderr(stderr_capture):
                await asyncio.wait_for(
                    asyncio.get_event_loop().run_in_executor(executor, exec, code_setup, student_code_vars, local_vars),
                    timeout=timeout,
                )

    except Exception as e:
        execution_error = e
        # TODO need to create a different message for setup code errors. This should result
        # in a different error message reported from the test case.

    if names_for_user_list is not None:
        for name_info in names_for_user_list:
            var_name = name_info["name"]

            if var_name in local_vars:
                # NOTE I think there might be issues with security with deepcopying certain
                # objects. If needed, we can prevent leaks here through serialization.
                student_code_vars[var_name] = deepcopy(local_vars[var_name])

    if execution_error is None:
        try:
            # Next, compile student code. Make sure to handle errors in this later
            # TODO have a better filename
            code_setup = compile(student_code, student_file_name, "exec")
            with redirect_stdout(stdout_capture), redirect_stderr(stderr_capture):
                await asyncio.wait_for(
                    asyncio.get_event_loop().run_in_executor(executor, exec, code_setup, student_code_vars, student_code_vars),
                    timeout=timeout,
                )

        except Exception as e:
            execution_error = e
            # TODO this traceback only shows the last line with the exception.
            # Would be better if we could give the full traceback within the student code
            # for example, if the student code calls a function that raises an exception,
            # we should show the full self-contained traceback including the function call.
            exception_traceback = traceback.format_exc(limit=-1)

    result_dict: ProcessStartResponse = {
        "status": ProcessStatusCode.SUCCESS if execution_error is None else ProcessStatusCode.EXCEPTION,
        "stdout": stdout_capture.getvalue(),
        "stderr": stderr_capture.getvalue(),
        "execution_error": type(execution_error).__name__ if execution_error else None,
        "execution_message": str(execution_error) if execution_error else None,
        "execution_traceback": str(exception_traceback),
    }

    return local_vars, student_code_vars, result_dict


async def handle_client(reader: asyncio.StreamReader, writer: asyncio.StreamWriter) -> None:
    """
    Reads lines from stdin asynchronously and responds on stdout.
    Mimics a simple server handling requests.
    """
    # try:
    #     json_message = json.loads(message)
    #     result = await asyncio.wait_for(
    #         asyncio.get_event_loop().run_in_executor(executor, _run_blocking_task, task_payload), timeout=timeout_seconds
    #     )
    #     # Example: send an acknowledgement back
    #     response_message = {"status": "received", "data": json_message}
    #     writer.write(json.dumps(response_message).encode("utf-8") + b"\n")  # Add newline for stream parsing
    #     await writer.drain()
    #     # ------------------------------------

    # except json.JSONDecodeError as e:
    #     error_response = {"status": "error", "message": f"Invalid JSON: {e}"}
    #     writer.write(json.dumps(error_response).encode("utf-8") + b"\n")
    #     await writer.drain()
    # except UnicodeDecodeError as e:
    #     error_response = {"status": "error", "message": f"Invalid UTF-8 encoding: {e}"}
    #     writer.write(json.dumps(error_response).encode("utf-8") + b"\n")
    #     await writer.drain()

    try:
        student_code_vars: None | dict = None
        local_vars: None | dict = None

        async for line_bytes in reader:
            line = line_bytes.decode().strip()
            if not line:  # Handle empty lines
                continue

            json_message = json.loads(line)

            msg_type = json_message.get("message_type")
            if msg_type == "start":
                start_json_message: ProcessStartRequest = json_message
                # Execute the student code for the first time and load
                # variables into the student_code_vars dictionary
                student_code = start_json_message["student_code"]
                student_file_name = start_json_message["student_file_name"]
                setup_code = start_json_message["setup_code"]
                initialization_timeout = start_json_message["initialization_timeout"]
                import_whitelist = start_json_message["import_whitelist"]
                import_blacklist = start_json_message["import_blacklist"]
                starting_vars = start_json_message["starting_vars"]
                builtin_whitelist = start_json_message["builtin_whitelist"]
                names_for_user_list = start_json_message["names_for_user_list"]

                populate_linecache(student_code, student_file_name)

                local_vars, student_code_vars, start_response = await student_code_runner(
                    setup_code=setup_code,
                    student_code=student_code,
                    student_file_name=student_file_name,
                    timeout=initialization_timeout,
                    import_whitelist=import_whitelist,
                    import_blacklist=import_blacklist,
                    starting_vars=starting_vars,
                    builtin_whitelist=builtin_whitelist,
                    names_for_user_list=names_for_user_list,
                )

                writer.write((json.dumps(start_response) + os.linesep).encode())

            elif msg_type == "query_setup":
                assert local_vars is not None
                query_setup_json_message: SetupQueryRequest = json_message

                var_to_query = query_setup_json_message["var"]
                if var_to_query in local_vars:
                    setup_query_response: SetupQueryResponse = {
                        "status": QueryStatusCode.SUCCESS,
                        "value_encoded": serialize_object_unsafe(local_vars[var_to_query]),
                    }
                else:
                    setup_query_response = {"status": QueryStatusCode.NOT_FOUND, "value_encoded": ""}

                writer.write((json.dumps(setup_query_response) + os.linesep).encode())

            elif msg_type == "query":
                assert student_code_vars is not None
                query_json_message: StudentQueryRequest = json_message

                var_to_query = query_json_message["var"]

                # Check if the variable exists in the student_code_vars
                if var_to_query in student_code_vars:
                    query_response: StudentQueryResponse = {
                        "status": QueryStatusCode.SUCCESS,
                        "value": to_json(student_code_vars[var_to_query]),
                    }
                else:
                    query_response = {"status": QueryStatusCode.NOT_FOUND, "value": ""}

                writer.write((json.dumps(query_response) + os.linesep).encode())

            elif msg_type == "query_function":
                assert student_code_vars is not None
                query_function_json_message: StudentFunctionRequest = json_message

                func_name = query_function_json_message["function_name"]
                args = deserialize_object_unsafe(query_function_json_message["args_encoded"])
                kwargs = deserialize_object_unsafe(query_function_json_message["kwargs_encoded"])
                query_timeout = query_function_json_message["query_timeout"]

                function_response = await student_function_runner(student_code_vars, func_name, query_timeout, args, kwargs)

                writer.write((json.dumps(function_response) + os.linesep).encode())

            # TODO handle cases of different payloads
            # The first payload should be student code
            if line.lower() == "exit":
                writer.write(("Goodbye!" + os.linesep).encode())
                await writer.drain()
                break  # Exit the loop and terminate the server

            # Simulate processing a request
            # response = f"Server processed: '{line.upper()}'\n"

            await writer.drain()  # Ensure the response is written to stdout

    except asyncio.CancelledError:
        writer.write((json.dumps({"status": "failure", "message": "Server was cancelled."}) + os.linesep).encode())
    except asyncio.TimeoutError:
        writer.write((json.dumps({"status": "failure", "message": "Student code timed out."}) + os.linesep).encode())
    except Exception as e:
        writer.write((json.dumps({"status": "failure", "message": f"An error occurred: {e}"}) + os.linesep).encode())
    finally:
        # It's good practice to close transports and writers
        # print("Closing server connections...")
        await writer.drain()  # Ensure all data is sent before closing
        writer.close()
        await writer.wait_closed()  # Wait for the writer to finish closing


async def main():
    """
    Starts the asynchronous socket server.
    """
    # Ensure ProactorEventLoop is used on Windows for robust socket operations
    if sys.platform == "win32":
        try:
            loop = asyncio.ProactorEventLoop()
            asyncio.set_event_loop(loop)
            print("Using ProactorEventLoop on Windows.", file=sys.stderr)
        except NotImplementedError:
            print("ProactorEventLoop not available, continuing with default loop.", file=sys.stderr)

    line_limit = 10 * 1024 * 1024  # 10 MB line limit to handle large messages

    # Start the server, binding to the specified host and port
    server = await asyncio.start_server(handle_client, HOST, 0, limit=line_limit)
    addr = server.sockets[0].getsockname()
    print(f"{addr[0]}, {addr[1]}", flush=True)

    async with server:
        # Run forever, or until the server is explicitly stopped
        await server.serve_forever()


if __name__ == "__main__":
    asyncio.run(main())



import numpy as np
import numpy.linalg as la


def solve_ratings(members, sale_metadata):
    sales = np.zeros(len(sale_metadata))
    theMatrix = np.zeros((len(members), len(sale_metadata)))
    for i in range(len(sale_metadata)):
        for j in range(len(sale_metadata[i]["hours"])):
            col = members.index(sale_metadata[i]["hours"][j][0])
            theMatrix[i][col] = sale_metadata[i]["hours"][j][1]
        sales[i] = sale_metadata[i]["units_sold"]
    ratings = la.solve(theMatrix, sales)
    highest_performing = members[np.argmax(ratings)]
    lowest_performing = members[np.argmin(ratings)]
    member_ratings = dict(zip(members, ratings))

    return member_ratings, highest_performing, lowest_performing



import base64
import builtins
import os
import pwd
from enum import StrEnum
from typing import Any
from typing import Literal
from typing import TypedDict

import dill
import pytest


class GradingOutputLevel(StrEnum):
    ExceptionName = "none"
    ExceptionMessage = "message"
    FullTraceback = "traceback"


class QueryStatusCode(StrEnum):
    """
    Status codes for a query operation.
    """

    SUCCESS = "success"
    NOT_FOUND = "not_found"


class FunctionStatusCode(StrEnum):
    """
    Status codes for a function execution.
    """

    SUCCESS = "success"
    EXCEPTION = "exception"
    TIMEOUT = "timeout"
    NOT_FOUND = "not_found"


class ProcessStatusCode(StrEnum):
    """
    Status codes for a process execution.
    """

    SUCCESS = "success"
    EXCEPTION = "exception"
    TIMEOUT = "timeout"
    NO_RESPONSE = "no_response"


# TODO use some inheritance on the query and response types


class NamesForUserInfo(TypedDict):
    name: str
    description: str
    type: str


class SetupQueryRequest(TypedDict):
    message_type: Literal["query_setup"]
    var: str


class SetupQueryResponse(TypedDict):
    status: QueryStatusCode
    value_encoded: str


# Variable query dict types
class StudentQueryRequest(TypedDict):
    message_type: Literal["query"]
    var: str
    query_timeout: float


class StudentQueryResponse(TypedDict):
    # This is meant to be deserialized into a Python object
    status: QueryStatusCode
    value: Any


# Function query dict types
class StudentFunctionRequest(TypedDict):
    message_type: Literal["query_function"]
    function_name: str
    args_encoded: str  # TODO add a stronger type for the input/output of the serialized function
    kwargs_encoded: str
    query_timeout: float


class StudentFunctionResponse(TypedDict):
    # This is meant to be deserialized into a Python object
    status: FunctionStatusCode
    value: Any
    stdout: str
    stderr: str
    exception_name: str | None
    exception_message: str | None
    traceback: str | None


# Process start dict types


class ProcessStartRequest(TypedDict):
    message_type: Literal["start"]
    student_code: str
    student_file_name: str
    setup_code: str | None
    initialization_timeout: float
    import_whitelist: list[str] | None
    import_blacklist: list[str] | None
    starting_vars: dict[str, Any] | None
    builtin_whitelist: list[str] | None
    names_for_user_list: list[NamesForUserInfo] | None


class ProcessStartResponse(TypedDict):
    status: ProcessStatusCode
    stdout: str
    stderr: str
    execution_error: str | None
    execution_message: str | None
    execution_traceback: str


def serialize_object_unsafe(obj: object) -> str:
    """
    Serializes an arbitrary Python object to a JSON string.
    The object is first serialized using dill, then base64 encoded.

    Returns:
        A JSON string representing the serialized object.
    """
    # 1. Serialize the object using dill
    dilled_bytes = dill.dumps(obj)

    # 2. Base64 encode the byte stream
    base64_encoded_bytes = base64.b64encode(dilled_bytes)

    # 3. Decode base64 bytes to a UTF-8 string for JSON storage
    return base64_encoded_bytes.decode("utf-8")


def deserialize_object_unsafe(base64_string: str) -> object:
    """
    Deserializes a Python object from a JSON string.

    The string is expected to contain a base64-encoded, dill-serialized
    object.

    Returns:
        The deserialized Python object.
    """

    # 1. Encode the base64 string back to bytes
    base64_encoded_bytes = base64_string.encode("utf-8")

    # 2. Base64 decode the bytes
    dilled_bytes = base64.b64decode(base64_encoded_bytes)

    # 3. Deserialize the object using dill
    return dill.loads(dilled_bytes)


def get_builtins(builtin_whitelist: list[str] | None) -> dict[str, Any]:
    """
    Returns a dictionary of safe built-in functions and exceptions.
    This is used to restrict the built-ins available in the student code execution environment.
    From https://github.com/zopefoundation/RestrictedPython/blob/master/src/RestrictedPython/Guards.py
    """
    final_builtins = {}

    _safe_names = (
        "__build_class__",
        "None",
        "False",
        "True",
        "abs",
        "bool",
        "bytes",
        "callable",
        "chr",
        "complex",
        "divmod",
        "float",
        "hash",
        "hex",
        "id",
        "int",
        "isinstance",
        "issubclass",
        "len",
        "oct",
        "ord",
        "pow",
        "print",
        "range",
        "repr",
        "round",
        "slice",
        "sorted",
        "str",
        "set",
        "list",
        "dict",
        "tuple",
        "zip",
        "enumerate",
        "min",
        "max",
        "map",
    )

    _safe_exceptions = (
        "ArithmeticError",
        "AssertionError",
        "AttributeError",
        "BaseException",
        "BufferError",
        "BytesWarning",
        "DeprecationWarning",
        "EOFError",
        "EnvironmentError",
        "Exception",
        "FloatingPointError",
        "FutureWarning",
        "GeneratorExit",
        "IOError",
        "ImportError",
        "ImportWarning",
        "IndentationError",
        "IndexError",
        "KeyError",
        "KeyboardInterrupt",
        "LookupError",
        "MemoryError",
        "NameError",
        "NotImplementedError",
        "OSError",
        "OverflowError",
        "PendingDeprecationWarning",
        "ReferenceError",
        "RuntimeError",
        "RuntimeWarning",
        "StopIteration",
        "SyntaxError",
        "SyntaxWarning",
        "SystemError",
        "SystemExit",
        "TabError",
        "TypeError",
        "UnboundLocalError",
        "UnicodeDecodeError",
        "UnicodeEncodeError",
        "UnicodeError",
        "UnicodeTranslateError",
        "UnicodeWarning",
        "UserWarning",
        "ValueError",
        "Warning",
        "ZeroDivisionError",
    )

    for name in _safe_names:
        final_builtins[name] = getattr(builtins, name)

    for name in _safe_exceptions:
        final_builtins[name] = getattr(builtins, name)

    # TODO raise exception if the name is not in the built-ins?
    if builtin_whitelist is not None:
        # Filter the built-ins based on the whitelist
        for name in builtin_whitelist:
            if name not in final_builtins:
                final_builtins[name] = getattr(builtins, name)

    return final_builtins


def drop_privileges(user_name: str) -> None:
    """Sets the process user and group to the specified non-root user."""
    try:
        # Get the UID and GID for the target user (e.g., 'nobody')
        pwnam = pwd.getpwnam(user_name)
        target_uid = pwnam.pw_uid
        target_gid = pwnam.pw_gid
    except KeyError:
        raise ValueError(f"User '{user_name}' not found.")

    # 1. Set the real and effective GID
    # NOTE: Set GID before UID, as changing UID may restrict GID changes
    os.setgid(target_gid)
    os.setegid(target_gid)

    # 2. Set the real and effective UID
    os.setuid(target_uid)
    os.seteuid(target_uid)


def get_output_level_marker(marker: pytest.Mark | None) -> GradingOutputLevel:
    if marker and marker.kwargs and "level" in marker.kwargs:
        try:
            # 2. Attempt to convert the marker value to the Enum member
            return GradingOutputLevel(marker.kwargs["level"])

        except ValueError as e:
            # 3. If conversion fails, the input value is invalid.
            # Pytest will treat this unhandled exception as a test failure.

            valid_levels = [level.value for level in GradingOutputLevel]

            raise ValueError(
                f"Invalid 'level' value '{marker.kwargs['level']}' in the 'output_level' marker. "
                f"Must be one of the following: {', '.join(valid_levels)}"
            ) from e

    return GradingOutputLevel.FullTraceback



import time

time.sleep(0.1)  # Simulate a long-running import

x = 5


def f_slow(x, *, y):
    # Simulate slow running function
    time.sleep(0.4)
    return x + y


def f_fast(x, *, y):
    # Simulate slow running function
    time.sleep(0.1)
    return x + y



import pytest

from pytest_pl_grader.fixture import FeedbackFixture
from pytest_pl_grader.fixture import StudentFixture


@pytest.mark.grading_data(name="initialization_timeout", points=2)
@pytest.mark.sandbox_timeout(0.05)
def test_query(sandbox: StudentFixture) -> None:
    assert sandbox.query("x") == 5


@pytest.mark.grading_data(name="function_timeout", points=2)
@pytest.mark.sandbox_timeout(0.5)
def test_query_slow(sandbox: StudentFixture, feedback: FeedbackFixture) -> None:
    assert sandbox.query("x") == 5
    assert sandbox.query_function("f_fast", 2, y=3, query_timeout=0.2) == 5
    feedback.set_score(0.1)
    assert sandbox.query_function("f_slow", 2, y=3, query_timeout=0.2) == 5


@pytest.mark.grading_data(name="function_timeout", points=2)
@pytest.mark.sandbox_timeout(0.5)
def test_query_fast(sandbox: StudentFixture, feedback: FeedbackFixture) -> None:
    assert sandbox.query("x") == 5
    assert sandbox.query_function("f_fast", 2, y=3, query_timeout=0.5) == 5
    feedback.set_score(0.1)
    assert sandbox.query_function("f_slow", 2, y=3, query_timeout=0.5) == 5



import sympy as sp


def analyze_vibration(x_hat, threshold):
    approximation = 7
    true_value = float((sp.sin(6 * x_hat) + 3 * sp.cos(x_hat) + 6).evalf())
    production_ready = approximation > threshold

    return approximation, true_value, production_ready



import numpy as np
import pytest

from pytest_pl_grader.fixture import DataFixture
from pytest_pl_grader.fixture import FeedbackFixture
from pytest_pl_grader.fixture import StudentFixture


def compute_fp_list(L: float, n: int, start: float, end: float) -> tuple[float, float, list[float]]:
    """
    Given:

    - L: the logarithm base 2 of the smallest positive floating point number
    - n: the number of bits in the mantissa
    - start: the smallest fp number to find (inclusive)
    - end: the largest fp number to find (inclusive)

    We assume base == 2

    Computes:

    - largest_fp: the largest positive floating point number
    - machine_epsilon: the machine epsilon
    - fp_lst: a list of all possible normalized floating point numbers in the range [start, end]
    """
    # L = np.log2(smallest_fp)

    # 1) largest fp
    U = -L
    largest_fp = 2 ** (U + 1) * (1 - 2 ** (-(n + 1)))

    # 2) machine epsilon
    machine_epsilon = 2 ** (-n)

    fp_lst = []

    # determine the exp range that covers [start, end]
    exp_start = np.floor(np.log2(start)).astype(int)
    exp_end = np.floor(np.log2(end)).astype(int)
    exp_min = L
    exp_max = -L

    # we only look through what is necessary, a simpler solution would be to loop through all possible exponents
    for exp in range(max(exp_start, exp_min), min(exp_end, exp_max) + 1):
        # find the first and last possible mantissas that fall in range
        mantissa_start = (start / (2**exp)) - 1
        mantissa_end = (end / (2**exp)) - 1
        mantissa_start_bits = max(0, np.ceil(mantissa_start * (2**n)).astype(int))
        mantissa_end_bits = min((2**n) - 1, np.floor(mantissa_end * (2**n)).astype(int))

        # loop through the possible mantissas
        for mantissa_bits in range(mantissa_start_bits, mantissa_end_bits + 1):
            mantissa = 1 + (mantissa_bits / (2**n))  # normalized mantissa
            value = mantissa * (2**exp)  # floating point value
            fp_lst.append(value)

    return largest_fp, machine_epsilon, sorted(fp_lst)


# TODO add a sandbox at the module level and use it in these tests


@pytest.mark.grading_data(name="largest_fp", points=1)
def test_0(sandbox: StudentFixture, feedback: FeedbackFixture, data_json: DataFixture) -> None:
    largest_fp_st = sandbox.query("largest_fp")

    assert largest_fp_st is not None, "largest_fp should not be None"

    assert isinstance(largest_fp_st, float), "largest_fp is not of correct type. It should be a float."

    largest_fp_ref, _, _ = compute_fp_list(
        data_json["params"]["L"],
        data_json["params"]["n"],
        data_json["params"]["start"],
        data_json["params"]["end"],
    )

    assert largest_fp_st == largest_fp_ref, "largest_fp has incorrect value."

    feedback.add_message("Good Job!")
    # feedback.set_score(1)

    # if self.ref.largest_fp == self.st.largest_fp:
    #     feedback.add_feedback("Good Job!")
    #     feedback.set_score(1)
    # else:
    #     feedback.add_feedback("largest_fp has incorrect value.")
    #     feedback.set_score(0)


@pytest.mark.grading_data(name="machine_epsilon", points=1)
def test_1(sandbox: StudentFixture, data_json: DataFixture, feedback: FeedbackFixture) -> None:
    machine_epsilon_st = sandbox.query("machine_epsilon")

    assert machine_epsilon_st is not None, "machine_epsilon should not be None"

    assert isinstance(machine_epsilon_st, float), "machine_epsilon is not of correct type. It should be a float."

    _, machine_epsilon_ref, _ = compute_fp_list(
        data_json["params"]["L"],
        data_json["params"]["n"],
        data_json["params"]["start"],
        data_json["params"]["end"],
    )

    assert machine_epsilon_st == machine_epsilon_ref, "machine_epsilon has incorrect value."

    feedback.add_message("Good Job!")

    # if self.ref.machine_epsilon == self.st.machine_epsilon:
    #     feedback.add_feedback("Good Job!")
    #     feedback.set_score(1)
    # else:
    #     feedback.add_feedback("machine_epsilon has incorrect value.")
    #     feedback.set_score(0)


@pytest.mark.grading_data(name="fp_lst", points=8)
def test_2(sandbox: StudentFixture, data_json: DataFixture, feedback: FeedbackFixture) -> None:
    fp_lst_st = sandbox.query("fp_lst")

    assert fp_lst_st is not None, "fp_lst should not be None"

    assert isinstance(fp_lst_st, list), "fp_lst is not of correct type. It should be a list."

    assert len(fp_lst_st) > 0, "fp_lst should not be empty"

    _, _, fp_lst_ref = compute_fp_list(
        data_json["params"]["L"],
        data_json["params"]["n"],
        data_json["params"]["start"],
        data_json["params"]["end"],
    )

    assert len(fp_lst_st) == len(fp_lst_ref), "fp_lst has incorrect length."

    # feedback.add_message("Good Job!")

    # if self.ref.fp_lst == self.st.fp_lst:
    #     feedback.add_feedback("Good Job!")
    #     feedback.set_score(1)

    # check length of list
    # if not feedback.check_list("fp_lst", fp_lst_ref, fp_lst_st):
    #     feedback.set_score(0)
    #     return

    # check list content within range
    start = fp_lst_ref[0]
    end = fp_lst_ref[-1]
    for st_val in fp_lst_st:
        assert start <= st_val <= end, "found value in fp_lst not within range"

    assert fp_lst_st == fp_lst_ref, "fp_lst has incorrect value."

    # # check list contents
    # for st_val, ref_val in zip(fp_lst_st, fp_lst_ref):
    #     if st_val != ref_val:
    #         feedback.add_feedback("inaccurate value found in fp_lst")
    #         feedback.set_score(0)
    #         return

    feedback.add_message("Good Job!")



import json
from io import StringIO
from typing import Any

import networkx as nx
import numpy as np
import pandas as pd
from plot_serializer.matplotlib.deserializer import deserialize_from_json
from plot_serializer.matplotlib.serializer import MatplotlibSerializer


def to_json(v: Any) -> Any:
    if isinstance(v, np.number):
        return {
            "_type": "np_scalar",
            "_concrete_type": type(v).__name__,
            "_value": str(v),
        }
    elif isinstance(v, np.bool_):
        return {
            "_type": "np_bool",
            "_value": str(v),
        }
    elif isinstance(v, np.ndarray):
        if np.isrealobj(v):
            return {"_type": "ndarray", "_value": v.tolist(), "_dtype": str(v.dtype)}
        elif np.iscomplexobj(v):
            return {
                "_type": "complex_ndarray",
                "_value": {"real": v.real.tolist(), "imag": v.imag.tolist()},
                "_dtype": str(v.dtype),
            }
    elif isinstance(v, pd.DataFrame):
        # The next lines of code are required to address the JSON table-orient
        # generating numeric keys instead of strings for an index sequence with
        # only numeric values (c.f. pandas-dev/pandas#46392)
        df_modified_names = v.copy()

        if df_modified_names.columns.dtype in (np.float64, np.int64):  # type: ignore
            df_modified_names.columns = df_modified_names.columns.astype("string")

        # For version 2 storing a data frame, we use the table orientation alongside of
        # enforcing a date format to allow for passing datetime and missing (`pd.NA`/`np.nan`) values
        # Details: https://pandas.pydata.org/docs/reference/api/pandas.read_json.html
        # Convert to JSON string with escape characters
        encoded_json_str_df = df_modified_names.to_json(orient="table", date_format="iso")
        # Export to native JSON structure
        pure_json_df = json.loads(encoded_json_str_df)

        return {"_type": "dataframe_v2", "_value": pure_json_df}
    elif isinstance(v, MatplotlibSerializer):
        return {"_type": "matplotlib_serializer", "_value": v.to_json()}
    else:
        return v


def _has_value_fields(v, fields: list[str]) -> bool:
    """Return True if all fields in the '_value' dictionary are present."""
    return "_value" in v and isinstance(v["_value"], dict) and all(field in v["_value"] for field in fields)


def from_json(v_json) -> Any:
    if isinstance(v_json, dict) and "_type" in v_json:
        if v_json["_type"] == "complex":
            if _has_value_fields(v_json, ["real", "imag"]):
                return complex(v_json["_value"]["real"], v_json["_value"]["imag"])
            else:
                raise ValueError("variable of type complex should have value with real and imaginary pair")
        elif v_json["_type"] == "np_bool":
            if "_value" in v_json:
                return np.bool_(v_json["_value"] == "True")
            else:
                raise ValueError("variable of type np_bool should have value")
        elif v_json["_type"] == "np_scalar":
            if "_concrete_type" in v_json and "_value" in v_json:
                return getattr(np, v_json["_concrete_type"])(v_json["_value"])
            else:
                raise ValueError(f"variable of type {v_json['_type']} needs both concrete type and value information")
        elif v_json["_type"] == "ndarray":
            if "_value" in v_json:
                if "_dtype" in v_json:
                    return np.array(v_json["_value"]).astype(v_json["_dtype"])
                else:
                    return np.array(v_json["_value"])
            else:
                raise ValueError("variable of type ndarray should have value")
        elif v_json["_type"] == "complex_ndarray":
            if _has_value_fields(v_json, ["real", "imag"]):
                if "_dtype" in v_json:
                    return (np.array(v_json["_value"]["real"]) + np.array(v_json["_value"]["imag"]) * 1j).astype(v_json["_dtype"])
                else:
                    return np.array(v_json["_value"]["real"]) + np.array(v_json["_value"]["imag"]) * 1j
            else:
                raise ValueError("variable of type complex_ndarray should have value with real and imaginary pair")
        elif v_json["_type"] == "dataframe_v2":
            # Convert native JSON back to a string representation so that
            # pandas read_json() can process it.
            value_str = StringIO(json.dumps(v_json["_value"]))
            return pd.read_json(value_str, orient="table")
        elif v_json["_type"] == "matplotlib_serializer":
            if "_value" in v_json:
                return deserialize_from_json(v_json["_value"])
            else:
                raise ValueError("variable of type matplotlib_serializer should have value")
        else:
            raise ValueError("variable has unknown type {}".format(v_json["_type"]))
    return v_json


def from_server_json(v_json: dict) -> Any:
    """
    Copy of the corresponding function from `prairielearn.conversion_utils` that is used to convert JSON serialized values from the server.

    https://github.com/PrairieLearn/PrairieLearn/blob/master/apps/prairielearn/python/prairielearn/conversion_utils.py
    """
    if isinstance(v_json, dict) and "_type" in v_json:
        if v_json["_type"] == "complex":
            if _has_value_fields(v_json, ["real", "imag"]):
                return complex(v_json["_value"]["real"], v_json["_value"]["imag"])
            else:
                raise ValueError("variable of type complex should have value with real and imaginary pair")
        elif v_json["_type"] == "np_scalar":
            if "_concrete_type" in v_json and "_value" in v_json:
                return getattr(np, v_json["_concrete_type"])(v_json["_value"])
            else:
                raise ValueError(f"variable of type {v_json['_type']} needs both concrete type and value information")
        elif v_json["_type"] == "ndarray":
            if "_value" in v_json:
                if "_dtype" in v_json:
                    return np.array(v_json["_value"]).astype(v_json["_dtype"])
                else:
                    return np.array(v_json["_value"])
            else:
                raise ValueError("variable of type ndarray should have value")
        elif v_json["_type"] == "complex_ndarray":
            if _has_value_fields(v_json, ["real", "imag"]):
                if "_dtype" in v_json:
                    return (np.array(v_json["_value"]["real"]) + np.array(v_json["_value"]["imag"]) * 1j).astype(v_json["_dtype"])
                else:
                    return np.array(v_json["_value"]["real"]) + np.array(v_json["_value"]["imag"]) * 1j
            else:
                raise ValueError("variable of type complex_ndarray should have value with real and imaginary pair")
        elif v_json["_type"] == "dataframe":
            if _has_value_fields(v_json, ["index", "columns", "data"]):
                val = v_json["_value"]
                return pd.DataFrame(index=val["index"], columns=val["columns"], data=val["data"])
            else:
                raise ValueError("variable of type dataframe should have value with index, columns, and data")
        elif v_json["_type"] == "dataframe_v2":
            # Convert native JSON back to a string representation so that
            # pandas read_json() can process it.
            value_str = StringIO(json.dumps(v_json["_value"]))
            return pd.read_json(value_str, orient="table")
        elif v_json["_type"] == "networkx_graph":
            return nx.adjacency_graph(v_json["_value"])
        else:
            raise ValueError(f"variable has unknown type {v_json['_type']}")
    raise ValueError(f"Expected a JSON object with a '_type' field, but got: {v_json}")



import numpy as np


def analyze_vibration(x_hat, threshold):
    approximation = 9 + 6 * x_hat - 3 * x_hat * x_hat / 2 - 216 * x_hat * x_hat * x_hat / 6
    true_value = np.sin(6 * x_hat) + 3 * np.cos(x_hat) + 6
    production_ready = approximation > threshold

    return approximation, true_value, production_ready
